/***** pam.c **************************************
 * Description: Functions for PAM package.
 *
 * Author: Bernhard Haubold, haubold@evolbio.mpg.de
 * File created on Fri Sep  1 11:40:22 2006.
 **************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "tab.h"
#include "pam.h"
#include "eprintf.h"

#define MAXLINE 1000
#define SIZE_CHARSET 256

Matrix *getMatrix(FILE *fp, char *fieldSep){
  char *str;
  int numFields, maxRow;
  int i;
  Matrix *matrix;
  double *row;

  matrix = (Matrix *)emalloc(sizeof(Matrix));
  matrix->numRows = 0;
  matrix->numCols = 0;
  matrix->comment = (char *)emalloc(0*sizeof(char));

  maxRow = 1;
  matrix->mat = (double **)emalloc(sizeof(double *));
  tabSetFieldSep(fieldSep);
  while((str=tabGetLine(fp)) != NULL ){
    if(str[0] != '#'){
      numFields = tabNfield();
      /* make sure that each line has the same number of fields */
      if(matrix->numCols)
	if(matrix->numCols != numFields)
	  eprintf("line %d: %d fields; line %d: %d fields\n",matrix->numRows,numFields,matrix->numRows-1,matrix->numCols);
      matrix->numRows++;
      if(matrix->numRows == maxRow){
	maxRow *= 2;
	matrix->mat = (double **)erealloc(matrix->mat,maxRow*sizeof(double *));
      }
      row = (double *)emalloc(numFields*sizeof(double));
      matrix->mat[matrix->numRows-1] = row;
      for(i=0;i<numFields;i++)
	row[i] = atof(tabField(i));
      matrix->numCols = numFields;
    }else{
      matrix->comment = erealloc(matrix->comment,strlen(matrix->comment)+strlen(str));
      strcat(matrix->comment,str);
    }
  }
  return matrix;
}

SubstitutionMatrix *duplicateSubstitutionMatrix(SubstitutionMatrix *sm1){
  int i, j;
  SubstitutionMatrix *sm2;
  
  /* allocate memory */
  sm2 = (SubstitutionMatrix *)emalloc(sizeof(SubstitutionMatrix));
  sm2->mat = (double **)emalloc(sm1->size*sizeof(double *));
  for(i=0;i<sm1->size;i++)
    sm2->mat[i] = (double *)emalloc(sm1->size*sizeof(double));
  sm2->dic = (int *)emalloc(sm1->size*sizeof(int));
  sm2->revdic = (char *)emalloc(sm1->size*sizeof(char));
  /* copy values */
  sm2->size = sm1->size;
  for(i=0;i<sm2->size;i++){
    sm2->dic[i] = sm1->dic[i];
    sm2->revdic[i] = sm1->revdic[i];
    for(j=0;j<sm2->size;j++)
      sm2->mat[i][j] = sm1->mat[i][j];
  }
  return sm2;
}

void outputMatrix(FILE *fp, SubstitutionMatrix *matrix, char *sep){
  int i, j, k, nBlanks, colWidth;
  double x;
  
  colWidth = 2;
  nBlanks = colWidth;
  for(i=0;i<nBlanks;i++)
    fprintf(fp," ");
  for(i=0;i<matrix->size;i++)
    fprintf(fp,"%6c%s",matrix->revdic[i],sep);
  fprintf(fp,"\n");
  for(i=0;i<matrix->size;i++){
    fprintf(fp,"%c",matrix->revdic[i]);
    if(sep[0] == ' '){
      x = matrix->mat[i][0];
      if(x < 1)
	x++;
      nBlanks = colWidth - (int)(log(x)/log(10));
      for(k=0;k<nBlanks;k++)
	fprintf(fp," ");
    }else
      fprintf(fp,"%s",sep);
    for(j=0;j<matrix->size;j++)
      if(j){
	if(sep[0] == ' '){
	  x = matrix->mat[i][j];
	  if(x < 1)
	    x++;
	  nBlanks = colWidth - (int)(log(x)/log(10));
	  for(k=0;k<nBlanks;k++)
	    fprintf(fp," ");
	  fprintf(fp,"%.3f",matrix->mat[i][j]);
	}else
	  fprintf(fp,"%s%.3f",sep,matrix->mat[i][j]);
      }else
	fprintf(fp,"%.3f",matrix->mat[i][j]);
    fprintf(fp,"\n");
  }
}

void outputMatrixInt(FILE *fp, SubstitutionMatrix *matrix){
  int i, j;

  fprintf(fp,"  ");
  for(i=0;i<matrix->size;i++)
    fprintf(fp,"%3c ",matrix->revdic[i]);
  fprintf(fp,"\n");
  for(i=0;i<matrix->size;i++){
    fprintf(fp,"%c ",matrix->revdic[i]);
    for(j=0;j<matrix->size;j++)
      if(j)
	fprintf(fp," %3.0f",matrix->mat[i][j]);
      else
	fprintf(fp,"%3.0f",matrix->mat[i][j]);
    fprintf(fp,"\n");
  }
}

SubstitutionMatrix *matrixMult(SubstitutionMatrix *matA, SubstitutionMatrix *matB){
  SubstitutionMatrix *matC;
  int i, j, k;

  matC = (SubstitutionMatrix *)emalloc(sizeof(SubstitutionMatrix));
  matC->size = matA->size;
  matC->mat = (double **)emalloc(matC->size*sizeof(double *));
  matC->revdic = matA->revdic;
  for(i=0;i<matC->size;i++){
    matC->mat[i] = (double *)emalloc(matC->size*sizeof(double));
    for(j=0;j<matC->size;j++){
      matC->mat[i][j] = 0.0;
      for(k=0;k<matA->size;k++)
	matC->mat[i][j] += matA->mat[i][k] * matB->mat[k][j];
    }
  }
  return matC;
}

/* readSubstitutionMatrix: read susbstitution matrix from file;
 * do not distinguish between upper and lower case letters
 */
SubstitutionMatrix *readSubstitutionMatrix(FILE *fp){
  char *line;
  char *num, c;
  int i, j, l, n;
  SubstitutionMatrix *sm;
  
  /* memory allocation */
  sm = (SubstitutionMatrix *)emalloc(sizeof(SubstitutionMatrix));
  num = (char *)emalloc(50*sizeof(char));
  line = (char *)emalloc(MAXLINE*sizeof(char));
  sm->dic = (int *)emalloc(SIZE_CHARSET*sizeof(int));
  sm->revdic = (char *)emalloc(SIZE_CHARSET*sizeof(char));

  while((line = fgets(line, MAXLINE, fp)) != NULL){
    if(strstr(line,"#") == NULL)
      break;
  }
  /* discover alphabet from first line of matrix */
  sm->size = 0;
  for(i=0;i<strlen(line);i++){
    c = line[i];
    if(!isspace(c)){
      sm->revdic[sm->size] = c;
      sm->dic[tolower(c)] = sm->size++;
      sm->dic[toupper(c)] = sm->dic[tolower(c)];
    }
  }
  sm->revdic = (char *)realloc(sm->revdic,(sm->size+1)*sizeof(char));
  sm->revdic[sm->size] = '\0';
  sm->mat = (double **)emalloc(sm->size*sizeof(double *));
  for(i=0;i<sm->size;i++){
    sm->mat[i] = (double *)emalloc(sm->size*sizeof(double));
  }
  /* loop over matrix body */
  for(i=0;i<sm->size;i++){
    line = fgets(line, MAXLINE, fp);
    l = 0;
    /* skip leading characters and blanks */
    while(isspace(line[l]) || isalpha(line[l]))
      l++;
    /* loop over elements in line */
    for(j=0;j<sm->size;j++){
      n = 0;
      while(!isspace(line[l])){
	num[n] = line[l];
	n++;
	l++;
      }

      num[n] = '\0';
      sm->mat[i][j] = atof(num);
      while(isspace(line[l]))
	l++;
    }
  }
  free(line);
  free(num);
  return sm;
}
